/*
 *Copyright (C) 1994-2000 The XFree86 Project, Inc. All Rights Reserved.
 *
 *Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 *"Software"), to deal in the Software without restriction, including
 *without limitation the rights to use, copy, modify, merge, publish,
 *distribute, sublicense, and/or sell copies of the Software, and to
 *permit persons to whom the Software is furnished to do so, subject to
 *the following conditions:
 *
 *The above copyright notice and this permission notice shall be
 *included in all copies or substantial portions of the Software.
 *
 *THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *NONINFRINGEMENT. IN NO EVENT SHALL THE XFREE86 PROJECT BE LIABLE FOR
 *ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 *CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *Except as contained in this notice, the name of the XFree86 Project
 *shall not be used in advertising or otherwise to promote the sale, use
 *or other dealings in this Software without prior written authorization
 *from the XFree86 Project.
 *
 * Authors:	David Fries <dfries@mail.win.org>
 */
/* $XFree86: xc/programs/Xserver/hw/xwin/winclipboard.h,v 1.3 2003/10/02 13:30:10 eich Exp $ */

/* This file is compiled into a DLL for windows with routines that will
 * intercept all keyboard keys when the enable is called.  The key events
 * will then be sent directly to the window passed in the enable call.
 * The routines will only operate when they are loaded in a DLL, and could
 * cause great greif it it is use improperly and no other applications
 * can get any keystrokes.
 */

//#include <process.h>
//#include <winuser.h>
//#include <iostream.h>

#include <stdio.h>
#include "win.h"
#include "keyboardhook.h"

#define KEYBOARD_HOOK_SYNC_EVENT WM_APP+32
#define KEYBOARD_HOOK_BYPASS_EVENT WM_APP+33

static BOOL relay = FALSE;
static HHOOK keyboardHook = NULL;
static HINSTANCE calling_instance = NULL;
static long calling_thread = 0;
static HWND calling_hWnd = NULL;

//#define DEBUG 1

static LRESULT CALLBACK
	keyboardHookRoutine(int code, WPARAM wParam, LPARAM lParam);

int enableKeyboardHook(HWND hWnd)
{
	if(relay)
	{
#ifdef DEBUG
		fprintf(stderr, "keyboard hook already in place\n");
#endif
		return 0;
	}
	calling_hWnd = hWnd;
	keyboardHook = SetWindowsHookEx(WH_KEYBOARD_LL,
		&keyboardHookRoutine, calling_instance, 0);
	if(keyboardHook)
	{
#ifdef DEBUG
		fprintf(stderr, "Keyboard hook successfully placed\n");
#endif
		relay = TRUE;
		return 1;
	}
	else
	{
#ifdef DEBUG
		fprintf(stderr, "Error installing keyboard hook\n");
		decodeGetLastErrorMessage(GetLastError());
#endif
		return 0;
	}
}

int disableKeyboardHook()
{
	if(UnhookWindowsHookEx(keyboardHook))
	{
#ifdef DEBUG
		fprintf(stderr, "Removing keyboard hook\n");
	}
	else
	{
		if(relay)
			fprintf(stderr, "Error removing keyboard hook\n");
		else
			fprintf(stderr, "Error removing keyboard hook, "
				"but that is expected\n");
		decodeGetLastErrorMessage(GetLastError());
#endif
	}
	relay = FALSE;
	calling_hWnd = NULL;
	return 0;
}

static BOOL window_key = FALSE;

static LRESULT CALLBACK
keyboardHookRoutine(int code, WPARAM wParam, LPARAM lParam)
{
	KBDLLHOOKSTRUCT * keyinfo = (KBDLLHOOKSTRUCT*)lParam;
	if( code >= 0 && relay && keyinfo->vkCode != 44)
	{
#ifdef DEBUG
		fprintf(stderr, "virtual key code %d Scan Code %d flags %d",
			keyinfo->vkCode, keyinfo->scanCode, keyinfo->flags);
		if(keyinfo->flags&1)
			fprintf(stderr, " extended ");
		else
			fprintf(stderr, " normal ");
		if(wParam == WM_KEYDOWN)
			fprintf(stderr, "Key down\n");
		else if(wParam == WM_KEYUP)
			fprintf(stderr, "Key up\n");
		else if(wParam == WM_SYSKEYDOWN)
			fprintf(stderr, "System Key down\n");
		else if(wParam == WM_SYSKEYUP)
			fprintf(stderr, "System Key up\n");
#endif
		// missing repeat count, and previous key state
		unsigned int repeat_count = 1;
		unsigned int paramOut = repeat_count | // dummy for now
			(keyinfo->scanCode<<16) | // scancode bits 0-16 to 16-23
			((keyinfo->flags&1)<<24) | // extended key 0 to 24
			((keyinfo->flags&32)<<24) | // alt down 5 to 29
						// previous state to 30
			((keyinfo->flags&128)<<24); // transition state 7 to 31
		/*
		PostThreadMessage(calling_thread, KEYBOARD_HOOK_SYNC_EVENT,
			keyinfo->vkCode, paramOut);
			*/
		if((keyinfo->scanCode == 91 && keyinfo->vkCode == 91) ||
			(keyinfo->scanCode == 91 && keyinfo->vkCode == 91))
			window_key = wParam == WM_KEYDOWN ||
				wParam == WM_SYSKEYDOWN;
	#if 0
		/* It would seem that Windows-200 and Windows-XP do things
		 * differently.  In Widows-XP this routine is called
		 * on down and up for Windows-U and the Utility Manager isn't
		 * called at all where Windows-2000 the Utility Manager is
		 * always called and we only get a up event for the U.
		 */
		if(keyinfo->scanCode == 22 && keyinfo->vkCode == 85 &&
			window_key && wParam == WM_KEYUP)
		{
			// The Window-U key doesn't send a down, but
			// only a up, so send the missing down.  We
			// should probably keep track of all the key
			// states so we don't send a down if there was
			// already a down that happened.
			PostMessage(calling_hWnd, WM_KEYDOWN, keyinfo->vkCode,
				paramOut);
		}
	#endif
		if(!PostMessage(calling_hWnd, wParam,
			keyinfo->vkCode, paramOut))
		{
			fprintf(stderr, "Warning PostThreadMessage failed, "
				"aborting key capture\n");
			disableKeyboardHook();
		}
		#if 0
		// 164 control, 162 alt, 46 delete, 13 enter
		if(keyinfo->vkCode == 164 || keyinfo->vkCode == 162 ||
			keyinfo->vkCode == 46 || keyinfo->vkCode == 13)
			return 1;
		#endif
		return 1;
	}
	if(keyinfo->vkCode == 44)
	{
		fprintf(stderr, "Disabling key capture\n");
		disableKeyboardHook();
	}
#ifdef DEBUG
	if(code < 0)
		fprintf(stderr, "Code < 0\n");
	if(!relay)
		fprintf(stderr, ":");
#endif
	return CallNextHookEx(keyboardHook, code, wParam, lParam);
}

void decodeGetLastErrorMessage(unsigned int error)
{
	LPVOID lpMsgBuf;
	FormatMessage( 
		FORMAT_MESSAGE_ALLOCATE_BUFFER | 
		FORMAT_MESSAGE_FROM_SYSTEM | 
		FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		error,
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
		(LPTSTR) &lpMsgBuf,
		0,
		NULL);

	fprintf(stderr, "Error %d : %s", error, (LPCSTR)lpMsgBuf);
	// Free the buffer.
	LocalFree( lpMsgBuf );
}

BOOL APIENTRY DllMain(HANDLE hModule, DWORD ul_reason_for_call,
	LPVOID lpReserved)
{
	switch (ul_reason_for_call)
	{
	case DLL_PROCESS_ATTACH:
		if(calling_instance)
		{
			fprintf(stderr, "Error keyboard hook is already  in "
				"use\n");
			return FALSE;
		}
		calling_instance = (HINSTANCE)hModule;
		calling_thread = GetCurrentThreadId();
		break;
	case DLL_PROCESS_DETACH:
	{
		disableKeyboardHook();
		calling_instance = NULL;
		calling_thread = 0;
	}
		break;
	case DLL_THREAD_ATTACH:
	case DLL_THREAD_DETACH:
		break;
	}
	return TRUE;
}
